#!/usr/bin/env python

import click

from libpipe import worker
from libpipe import __version__


t_file = click.Path(exists=True, dir_okay=False)


@click.group()
@click.version_option(__version__)
def main():
    ''' Nodes utilities ...'''


@main.command('run_all')
@click.argument('cmd_list', type=t_file)
@click.option('--nodes_string', '-n', help='Range of nodes to use', type=str, default='localhost')
@click.option('--max_concurrent', '-m', help='Maximum concurrent tasks on a node', type=int, default=5)
@click.option('--env_file', '-e', help='Environment source file', type=t_file)
@click.option('--dry_run', '-d', help='Dry run, only print commands', is_flag=True)
def run_all(cmd_list, nodes_string, max_concurrent, env_file, dry_run):
    ''' Run all command listed in the file CMD_LIST.'''
    pool = worker.get_worker_pool('Run All', nodes_string, max_concurrent=max_concurrent,
                                  env_file=env_file, dry_run=dry_run, debug=dry_run)

    for cmd in open(cmd_list).readlines():
        cmd = cmd.strip()
        pool.add(cmd)

    pool.execute()


@main.command('run_on_nodes')
@click.argument('cmd', nargs=-1)
@click.option('--nodes_string', '-n', help='Range of nodes to use', type=str, default='localhost')
@click.option('--env_file', '-e', help='Environment source file', type=t_file)
@click.option('--dry_run', '-d', help='Dry run, only print commands', is_flag=True)
def run_on_all_nodes(cmd, nodes_string, env_file, dry_run):
    ''' Run CMD on all nodes specified in argument.'''
    pool = worker.get_worker_pool('Run All', nodes_string, max_concurrent=2,
                                  env_file=env_file, dry_run=dry_run, debug=dry_run)

    cmd = ' '.join(cmd)

    for host in worker.get_hosts(nodes_string):
        pool.add(cmd, run_on_host=host)

    pool.execute()



# Multi-process, multi-nodes process executors
#
# Authors: F.Mertens

import os
import re
import time
import socket
import signal
import getpass
import asyncio
import datetime
import itertools
import collections
from enum import Enum

import asyncssh
import progressbar
from click import style, secho


localhost_shortname = socket.gethostname().split('.', 1)[0]


def kill_remote(host, user, pid):
    os.system(f'ssh {user}@{host} "kill -TERM -{pid}"')


def n_digits(i):
    return len(str(i))


def expend_num_ranges(s):
    r = re.split(r'\[([0-9-,]+)\]', s)
    for i in range(1, len(r), 2):
        if ',' in r[i]:
            r[i] = r[i].split(',')
        elif '-' in r[i]:
            s, e = r[i].split('-')
            n = max(n_digits(s), n_digits(e))
            r[i] = [str(k).rjust(n, '0') for k in range(int(s), int(e) + 1)]
    for i in range(0, len(r), 2):
        r[i] = [r[i]]
    for el in itertools.product(*r):
        yield ''.join(el)


def get_hosts(host_string):
    hosts = list(set(host for k in host_string.split(',') for host in expend_num_ranges(k)))
    return [localhost_shortname if h == 'localhost' else h for h in hosts]


def get_worker_pool(name, nodes='localhost', max_concurrent=4, env_file=None, max_time=None,
                    debug=False, dry_run=False):
    hosts = get_hosts(nodes)
    return WorkerPool(hosts, name=name, max_tasks_per_worker=max_concurrent,
                      env_source_file=env_file, max_time=max_time, debug=debug, dry_run=dry_run)


class TaskStatus(Enum):
    WAITING = 1
    RUNNING = 2
    CANCELED = 3
    SUCCESS = 4
    ERROR = 5


class WorkerStatus(Enum):
    IDLE = 1
    COMMITED = 2
    RUNNING = 3
    ERROR = 4


class Process(object):

    def __init__(self, client, process):
        self.client = client
        self.process = process


class Task(object):
    """Represent a task executed by the worker pool

    Attributes:
        name (str): Name of the task (set by the worker pool).
        command (str): Command
        output_file (str): Optional log file. None if not set.
        returncode (int): Return code. None if task was not executed.
    """

    def __init__(self, name, command, output_file=None, done_callback=None, run_on_host=None, n_try=3):
        self.name = name
        self.command = command
        self.output_file = output_file
        self.fd = None
        self.returncode = None
        self.n_try = 0
        self.done_callback = done_callback
        self.run_on_host = run_on_host
        self.status = TaskStatus.WAITING
        self.process = None
        self.n_try = n_try
        self.timed_out = False

    def init_log(self):
        if self.output_file is not None:
            self.fd = open(self.output_file, 'w')
            self.fd.write(f'# Logging starting at {datetime.datetime.now()}\n')
            self.fd.write(f'# Input command: {self.command}\n\n')

    def set_process(self, process):
        self.process = process

    async def terminate(self):
        if self.process is not None:
            if hasattr(self.process, 'pid'):
                if self.process.client.host is not 'local':
                    self.log(f'Killing local process PID {self.process.pid}\n', 'local', err=True)
                    os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
                else:
                    self.log(f'Killing remote process PID {self.process.pid}\n', self.process.client.host, err=True)
                    kill_remote(self.process.client.host, self.process.client.user, self.process.pid)

    def log(self, line, host, err=False):
        if isinstance(line, bytes):
            line = line.decode()
        if self.fd is not None:
            self.fd.write(line)
        line = line.strip()
        if err:
            line = style(line, fg='red')
        print(f'[{self.name}:{host}] {line}')

    async def log_stream(self, stream, host, err=False):
        async for line in stream:
            self.log(line, host, err=err)

    def close_log(self):
        if self.fd is not None:
            try:
                self.fd.write(f'\n\n# Logging stopped at {datetime.datetime.now()}\n')
                self.fd.close()
            except Exception:
                print(f'Error closing log file for task {self.name}. Ignoring.')

    async def run(self, client_pool):
        self.init_log()

        for i in range(self.n_try):
            self.timed_out = False

            worker = await client_pool.get_worker(self.run_on_host)
            self.status = TaskStatus.RUNNING

            ret_status, err_msg = await worker.run(self)

            if err_msg:
                self.log(err_msg, worker.client.host, err=True)

            if ret_status and not self.timed_out:
                self.status = TaskStatus.SUCCESS
                break

            self.status = TaskStatus.ERROR
            if i + 1 < self.n_try:
                self.log(f'Will retry {i + 1} / {self.n_try} ...\n', worker.client.host, err=True)

        self.close_log()


class SSHClient(asyncssh.SSHClient):

    def __init__(self):
        self.connected = False

    def connection_made(self, conn):
        self.connected = True
        print('Connection made to %s.' % conn.get_extra_info('peername')[0])

    def connection_lost(self, exc):
        if exc:
            print('SSH client error: ' + str(exc))
            raise exc
        self.connected = False


class Client(object):

    def __init__(self, host, user=None, force_sync=True, password=None):
        self.host = host
        self.user = user
        self.password = password
        self.conn = None
        self.client = None
        self.starting = False
        self.started = asyncio.Event()
        self.closing = False
        self.force_sync = force_sync
        self.creating_session = asyncio.Lock()

    async def start(self):
        if not self.starting:
            self.starting = True
            print(f'Starting client {self.host} ...')
            try:
                self.conn, self.client = await asyncssh.create_connection(SSHClient, self.host, username=self.user,
                                                                          password=self.password)
            except Exception:
                raise
            finally:
                self.started.set()

    async def execute(self, task):
        if not self.connected():
            raise ConnectionError('Client not connected')

        async with self.creating_session:
            process = await self.conn.create_process('echo $$;' + task.command)
        task.set_process(Process(self, process))

        process.pid = await process.stdout.readline()

        await task.log_stream(process.stdout, self.host)
        await task.log_stream(process.stderr, self.host, err=True)

        process.channel.close()
        task.returncode = process.returncode

        if self.force_sync:
            await self.conn.run('sync')
            await asyncio.sleep(1)

    def close(self):
        if self.conn is not None and not self.closing:
            self.closing = True
            self.conn.close()
            self.conn = None

    def connected(self):
        return self.client is not None and self.client.connected


class LocalClient(object):

    def __init__(self, host='local', force_sync=True):
        self.host = host
        self.started = asyncio.Event()
        self.force_sync = force_sync

    async def start(self):
        self.started.set()

    async def execute(self, task):
        process = await asyncio.create_subprocess_shell(task.command, stdout=asyncio.subprocess.PIPE,
                                                        stderr=asyncio.subprocess.PIPE, preexec_fn=os.setpgrp)
        task.set_process(process)

        await task.log_stream(process.stdout, self.host)
        await task.log_stream(process.stderr, self.host, err=True)

        task.returncode = await process.wait()

        if self.force_sync:
            os.sync()
            await asyncio.sleep(1)

    def connected(self):
        return self.started.is_set()

    def close(self):
        # nothing to do
        pass


class Worker(object):

    def __init__(self, client, name):
        self.client = client
        self.name = name
        self.status = WorkerStatus.IDLE
        self.execute_start_time = None
        self.running_task = None

    async def run(self, task):
        self.running_task = task
        self.execute_start_time = time.time()
        self.status = WorkerStatus.RUNNING
        err_msg = ''

        try:
            await self.client.start()
        except Exception as exc:
            err_msg = f'Error starting client: {str(exc)}\n'
            self.status = WorkerStatus.ERROR
            return False, err_msg

        await self.client.started.wait()

        try:
            await self.client.execute(task)
        except ConnectionError as exc:
            err_msg = f'Error connecting client: {str(exc)}\n'
            self.status = WorkerStatus.ERROR
            return False, err_msg
        except Exception as exc:
            err_msg = f'Error executing task {self.name}: {str(exc)}\n'
            self.status = WorkerStatus.IDLE
            return False, err_msg

        if task.done_callback is not None:
            try:
                task.done_callback()
            except Exception as exc:
                err_msg = f'Error executing return callback of task {self.name}: {str(exc)}\n'

        self.status = WorkerStatus.IDLE

        return True, err_msg

    def done(self):
        self.status = WorkerStatus.IDLE


class WorkerPool(object):

    def __init__(self, hosts, name='Worker', max_tasks_per_worker=4,
                 env_source_file=None, user=None, max_time=None, force_sync=False,
                 debug=False, dry_run=False, password=None):
        """Initiate a worker pool.

        Args:
            name (str): Name of the worker pool.
            hosts (list): List of host names, reachable via ssh (if not local host).
            max_tasks_per_worker (int, optional): Maximum number of tasks to execute concurrently on an host.
            env_source_file (str, optional): Name of a file to source before executing a task.
            user (str, optional): User to connect
        """
        if asyncio.get_event_loop().is_closed():
            asyncio.set_event_loop(asyncio.new_event_loop())

        if user is None:
            user = getpass.getuser()

        self.workers = []
        self.tasks = []
        self.env_file = env_source_file
        self.max_time = max_time
        self.name = name
        self.debug = debug
        self.dry_run = dry_run
        self.hosts = hosts
        self.futures = None
        self.clients = []
        self.active_hosts = hosts

        for host in self.hosts:
            if host == localhost_shortname:
                client = LocalClient(localhost_shortname, force_sync=force_sync)
            else:
                client = Client(host, user, force_sync=force_sync, password=password)

            for i in range(max_tasks_per_worker):
                worker = Worker(client, f'{client.host}:{i}')
                self.workers.append(worker)

        widgets = [
            f"{style(self.name, bold=True)}: ", progressbar.Percentage(), ' (',
            progressbar.SimpleProgress(), ')'
            ' ', progressbar.Bar(marker='|', left='[', right=']'),
            ' ', progressbar.Timer(),
            ' ', progressbar.ETA(),
        ]

        self.pbar = progressbar.ProgressBar(redirect_stdout=True, widgets=widgets)

    async def get_worker(self, host=None):
        while True:
            for worker in self.workers:
                if host is not None and worker.client.host != host:
                    if host not in self.active_hosts:
                        print(f'Requested run_on_host "{host}" not available anymore. Releasing constraint ...')
                        host = None
                    continue
                if worker.status is WorkerStatus.IDLE:
                    worker.status = WorkerStatus.COMMITED
                    return worker
            await asyncio.sleep(0.1)

    def add(self, command, name=None, output_file=None, done_callback=None, run_on_host=None, n_try=3):
        """Add a command to execute by a worker in the pool. Optionally output result in output_file.

        Args:
            command (str): Command to execute.
            name (str, optional): Name of the task. If not set it will be set by the worker pool
            output_file (str, optional): Optional filename to log output into.
        """
        if run_on_host is not None and run_on_host not in self.hosts:
            print(f'Requested run_on_host "{run_on_host}" is not in the hosts list. Releasing constraint ...')

        if self.env_file:
            command = "sh -c '. %s; %s'" % (self.env_file, command)
        if name is None:
            name = f'T{len(self.tasks)}'

        if self.debug:
            print(command)
        if self.dry_run:
            return

        self.tasks.append(Task(name, command, output_file=output_file, done_callback=done_callback,
                               run_on_host=run_on_host, n_try=n_try))

    def get_all_tasks(self):
        tasks = collections.defaultdict(list)
        for t in self.tasks:
            tasks[t.status].append(t)
        return tasks

    async def _monitor_queue(self, update_interval=1):
        try:
            n_todo = len(self.tasks)
            while n_todo > 0:
                await asyncio.sleep(update_interval)
                tasks = self.get_all_tasks()
                n_done = len(tasks[TaskStatus.SUCCESS])
                n_todo = len(tasks[TaskStatus.WAITING]) + len(tasks[TaskStatus.RUNNING])
                self.pbar.update(n_done)

                self.active_hosts = set(w.client.host for w in self.workers if w.status != WorkerStatus.ERROR)
                n_active_workers = 0

                for worker in self.workers:
                    n_active_workers += worker.status != WorkerStatus.ERROR

                    if worker.status == WorkerStatus.RUNNING and self.max_time is not None:
                        running_time = time.time() - worker.execute_start_time
                        if running_time > (self.max_time + update_interval):
                            task = worker.running_task
                            task.timed_out = True
                            task.log(f'Task timeout: waited for {running_time:.2f} s', worker.client.host, err=True)
                            task.log(f'Command was: {task.command}', worker.client.host, err=True)
                            await task.terminate()

                if n_active_workers == 0:
                    secho("No active workers, canceling all remaining tasks ...", fg='red')
                    for future in self.futures[::-1]:
                        future.cancel()

        except asyncio.CancelledError:
            pass

    async def process_tasks(self):
        self.pbar.start(max_value=len(self.tasks))

        self.futures = [asyncio.ensure_future(task.run(self)) for task in self.tasks]
        self.futures.append(asyncio.ensure_future(self._monitor_queue()))

        await asyncio.gather(*self.futures, return_exceptions=False)

    def execute(self):
        if not self.tasks:
            self.pbar.finish()
            return [], []

        loop = asyncio.get_event_loop()

        try:
            loop.run_until_complete(self.process_tasks())
        except asyncio.CancelledError:
            secho(f'Error: one (or more) task(s) have been canceled', fg='red')
        finally:
            for client in self.clients:
                client.close()

        tasks = self.get_all_tasks()
        n_success = len(tasks[TaskStatus.SUCCESS])

        self.pbar.update(n_success)
        self.pbar.finish(dirty=n_success != len(self.tasks))

        return tasks[TaskStatus.SUCCESS], tasks[TaskStatus.ERROR] + tasks[TaskStatus.CANCELED] \
            + tasks[TaskStatus.WAITING] + tasks[TaskStatus.RUNNING]



import os
import pathlib
from unittest import mock

from libpipe import settings


class Settings(settings.BaseSettings):

    DEFAULT_SETTINGS = pathlib.Path(__file__).absolute().parent / 'settings_test.toml'


def test_settings(tmp_path):
    c = Settings(None, {'vis': {'umin': 20, 'umax': 40, 'list': ['a', 'b']}, 'data_path': 'some_path'})

    assert c.vis.umin == 20
    assert c.vis.umax == 40
    assert c.data_path == 'some_path'
    assert c['vis'] == {'umin': 20, 'umax': 40, 'list': ['a', 'b']}
    assert c.vis.list == ['a', 'b']
    assert c.vis.list[1] == 'b'

    c.image = {'lmin': 100, 'data_col': 'DATA'}
    c.obs_ids = [1, 2, 3, 4]

    assert c.image.lmin == 100
    assert c.image.data_col == 'DATA'
    assert c.obs_ids == [1, 2, 3, 4]
    assert c.obs_ids[0] == 1

    s = '''
    db_path = 'some_other_path'

    [vis]
    list = ['c', 'd']

    [worker]
    nodes = 'node1[0-1][1-5]'
    concurrent = 2
    '''

    filename = tmp_path / 'test.toml'
    with open(str(filename), 'w') as f:
        f.write(s)

    c2 = Settings.load(str(filename))

    assert c2.vis.list == ['c', 'd']
    assert c2.worker.nodes == 'node1[0-1][1-5]'

    c3 = c2.copy()
    c3.worker.concurrent = 3
    c3.new = 'a'

    assert c2.worker.concurrent == 2
    assert c3.worker.concurrent == 3
    assert hasattr(c3, 'new')
    assert not hasattr(c2, 'new')

    c4 = c + c2

    assert c4.vis.umin == 20
    assert c4.vis.umax == 40
    assert c4.vis.list == ['c', 'd']
    assert c4.worker.nodes == 'node1[0-1][1-5]'
    assert c4.worker.concurrent == 2

    c5 = c + c2 + c3

    assert c5.worker.concurrent == 3
    assert c5.new == 'a'

    c = Settings.load_from_string(s)

    assert c.vis.list == ['c', 'd']
    assert c.worker.nodes == 'node1[0-1][1-5]'

    c = Settings.load_from_string(['db_path="a"', 'vis.c=2'])

    assert c.db_path == 'a'
    assert c.vis.c == 2


def test_default_settings(tmp_path):
    s = '''
    db_path = 'some_path'
    default_settings = ''

    [vis_cube]
    umax = 200
    win_fct = 'hann'

    [worker]
    nodes = 'node1[0-1][1-5]'
    concurrent = 2
    '''

    filename = tmp_path / 'config.toml'
    with open(str(filename), 'w') as f:
        f.write(s)

    c = Settings.load_with_defaults(filename)

    assert c.db_path == 'some_path'
    assert c.default_settings == ''
    assert c.vis_cube.umax == 200
    assert c.vis_cube.win_fct == 'hann'
    assert c.worker.nodes == 'node1[0-1][1-5]'
    assert c.worker.concurrent == 2
    assert c.data_dir == '.'
    assert c.vis_cube.umin == 50

    s = f'''
    db_path = 'some_other_path'
    default_settings = '{str(filename)}'

    [vis_cube]
    umax = 400

    [image]
    umax = 5000

    [worker]
    concurrent = 2
    '''

    filename2 = tmp_path / 'config2.toml'
    with open(str(filename2), 'w') as f:
        f.write(s)

    with mock.patch.object(Settings, 'DEFAULT_SETTINGS', filename):
        c = Settings.load_with_defaults(filename2)

        assert c.db_path == 'some_other_path'
        assert os.path.abspath(c.default_settings) == os.path.abspath(filename)
        assert c.vis_cube.umax == 400
        assert c.vis_cube.win_fct == 'hann'
        assert c.worker.nodes == 'node1[0-1][1-5]'
        assert c.worker.concurrent == 2
        assert c.image.umax == 5000

        c.save(strip_parents_keys=True)

        c = Settings.load_with_defaults(filename2)

        assert c.db_path == 'some_other_path'
        assert os.path.abspath(c.default_settings) == os.path.abspath(filename)
        assert c.vis_cube.umax == 400
        assert c.vis_cube.win_fct == 'hann'
        assert c.worker.nodes == 'node1[0-1][1-5]'
        assert c.worker.concurrent == 2
        assert c.image.umax == 5000

        c = Settings.load(filename2)

        print(open(filename2).readlines())
        assert c.db_path == 'some_other_path'
        assert os.path.abspath(c.default_settings) == os.path.abspath(filename)
        assert c.image.umax == 5000
        assert c.vis_cube.umax == 400
        assert not hasattr(c, 'worker')
        assert not hasattr(c.image, 'win_fct')


def test_duplicate(tmp_path):
    s = '''
    db_path = 'some_path'
    default_settings = ''

    [worker]
    nodes = 'node1[0-1][1-5]'
    concurrent = 2
    some_parset = 'config/some_file.parset'
    '''
    os.chdir(tmp_path)

    filename = tmp_path / 'config.toml'
    with open(str(filename), 'w') as f:
        f.write(s)

    os.makedirs(tmp_path / 'config')

    (tmp_path / 'config' / 'some_file.parset').touch()

    s = Settings.load_with_defaults(filename)

    new_file = tmp_path / 'new_config.toml'

    ns = s.duplicate(new_file)

    assert ns.db_path == 'some_path'
    assert ns.worker.some_parset == 'config/some_file.parset'
    assert not (tmp_path / 'new_dir' / 'config' / 'some_file.parset').is_file()

    new_file = tmp_path / 'new_dir' / 'new_config.toml'

    ns = s.duplicate(new_file, copy_parset_files=False)

    assert ns.db_path == 'some_path'
    assert ns.worker.some_parset == str(tmp_path / 'config' / 'some_file.parset')

    new_file = tmp_path / 'new_dir' / 'new_config.toml'

    ns = s.duplicate(new_file, copy_parset_files=True)

    assert ns.db_path == 'some_path'
    assert ns.worker.some_parset == 'config/some_file.parset'
    assert (tmp_path / 'new_dir' / 'config' / 'some_file.parset').is_file()

    ns = s.duplicate(new_file, update_relative_path=True, **{'db_path': 'added_path', 'worker': {'concurrent': 4}})

    assert ns.db_path == 'added_path'
    assert ns.worker.concurrent == 4


def test_get_path(tmp_path):
    c = Settings(None, {'data_dir': 'a'})
    assert c.get_path('data_dir') == str(pathlib.Path.cwd() / 'a')

    c = Settings('/some/file.toml', {'data_dir': 'a'})
    assert c.get_path('data_dir') == '/some/a'

    c = Settings(None, {'data_dir': '/a'})
    assert c.get_path('data_dir') == '/a'

    c = Settings('/some/file.toml', {'data_dir': '/a'})
    assert c.get_path('data_dir') == '/a'

    c = Settings.load_from_string('a.b.data_dir = "/some_dir"')
    assert c.get_path('a.b.data_dir') == '/some_dir'

    c = Settings.load_from_string('a.b.data_dir = "some_dir"')
    assert c.get_path('a.b.data_dir') == str(pathlib.Path.cwd() / 'some_dir')

    c = Settings.load_from_string('a.b.c.data_dir = "/some_dir"')
    assert c.a.b.get_path('c.data_dir') == '/some_dir'


def test_default_type(tmp_path):
    s = '''
    [type1]
    a = 200
    b = 'h'

    [type2]
    c = [1, 2, 3]
    d.k = 2

    [type3]
    pa.b = 3
    pa.c = '2'
    v = [1]

    [type4]
    a = 0

    [type5]
    u = 0
    '''

    filename = tmp_path / 'default.toml'
    with open(str(filename), 'w') as f:
        f.write(s)

    s = '''
    [action1]
    type = 'type1'
    a = 100

    [action2]
    type = 'type1'
    c = 200

    [action3]
    type = 'type2'
    c = [2, 3, 1, 2]
    d = 4

    [type3]
    pa.a = 1
    pa.c = '5'

    [type5]

    '''

    filename2 = tmp_path / 'config.toml'
    with open(str(filename2), 'w') as f:
        f.write(s)

    with mock.patch.object(Settings, 'DEFAULT_SETTINGS', filename):
        c = Settings.load_with_defaults(filename2, with_type_keyword=True)

    assert c.action1.a == 100
    assert c.action1.b == 'h'
    assert c.action2.a == 200
    assert c.action2.c == 200
    assert c.action3.d == 4
    assert c.action3.c == [2, 3, 1, 2]
    assert c.type3.pa.a == 1
    assert c.type3.pa.b == 3
    assert c.type3.pa.c == '5'
    assert c.type3.v == [1]
    assert 'type4' not in c
    assert 'type5' in c



# Handle setting
#
# Author: F. Mertens

import os
import shutil
from collections import OrderedDict

import toml

from . import attrmap, futils


class BaseSettings(attrmap.AttrMap):

    DEFAULT_SETTINGS = None

    def __init__(self, file, d):
        attrmap.AttrMap.__init__(self, d)
        self.set_file(file)

    def _build(self, obj):
        if isinstance(obj, attrmap.Mapping):
            obj = self.__class__(self.get_file(), obj)

        return obj

    def set_file(self, file):
        self._setattr('_file', os.path.abspath(file) if file else file)

    def get_file(self):
        return self._file

    def get_path(self, key):
        path = self.get(key)
        if path:
            path = os.path.expanduser(path)
            if not os.path.isabs(path):
                if self.get_file():
                    path = os.path.join(os.path.dirname(self.get_file()), path)
                else:
                    path = os.path.abspath(path)
        return path

    def validate_keys(self, max_depth=2):
        sd = self.get_defaults()
        invalid_keys = set(self.all_keys()) - set(sd.all_keys())
        if invalid_keys and min([len(k.split('.')) for k in invalid_keys]) <= max_depth:
            print(f'Warning: invalid keys: {",".join(invalid_keys)}')
            return False
        return True

    @classmethod
    def load(cls, file, check_args=True):
        s = cls(file, toml.load(file, _dict=OrderedDict))
        if check_args:
            s.validate_keys()

        return s

    @classmethod
    def load_with_defaults(cls, file, with_type_keyword=False, check_args=True):
        c_file = cls.load(file, check_args=check_args)
        c_default = cls.get_defaults()

        if 'default_settings' in c_file and c_file.default_settings:
            c_default += cls.load(c_file.get_path('default_settings'), check_args=check_args)

        if not with_type_keyword:
            s = c_default + c_file
        else:
            for k, v in c_file.items():
                if isinstance(v, dict):
                    if 'type' in v.keys():
                        t = v['type']
                    else:
                        t = k
                    if t not in c_default.keys():
                        print(f'Warning: no default for section {k} of type {t}')
                        continue
                    t_default = c_default[t]
                    c_file[k] = attrmap.AttrMap(t_default) + attrmap.AttrMap(v)
            s = c_file

        s.set_file(file)
        return s

    @classmethod
    def get_defaults(cls):
        return cls.load(cls.DEFAULT_SETTINGS, check_args=False)

    def get_root_settings_file(self):
        s = self
        while s.default_settings:
            s = self.load(s.get_path('default_settings'), check_args=False)
        return s.get_file()

    @classmethod
    def load_from_string(cls, s, check_args=True):
        if isinstance(s, (list, tuple)):
            s = '\n'.join(s)
        d = toml.loads(s)

        s = cls(None, d)
        if check_args:
            s.validate_keys()

        return s

    def save(self, file=None, strip_parents_keys=False):
        if not file:
            file = self.get_file()
        if not file:
            print('Error: No target filename to save settings.')
        else:
            m = self._mapping.copy()
            if strip_parents_keys and os.path.isfile(self.get_path('default_settings')):
                s_parent = self.__class__.load(self.default_settings)
                m = m - s_parent
            with open(file, mode='w') as f:
                toml.dump(dict(m), f)

    def duplicate(self, out_file, copy_parset_files=False, strip_parents_keys=False, **config_modifiers):
        s_out = self.copy()
        s_out.set_file(out_file)
        s_out['default_settings'] = ''

        cur_dir = os.path.dirname(self.get_file()) if self.get_file() else os.getcwd()
        new_dir = os.path.dirname(s_out.get_file()) if s_out.get_file() else os.getcwd()

        if new_dir and not os.path.exists(new_dir):
            os.makedirs(new_dir)

        if cur_dir != new_dir:
            if not os.path.isabs(s_out.data_dir):
                s_out['data_dir'] = os.path.join(cur_dir, s_out.data_dir)
            for p, k, v in s_out.all_items():
                if isinstance(v, str) and v.endswith('.parset') and not os.path.isabs(v):
                    if copy_parset_files:
                        futils.mkdir(os.path.join(new_dir, os.path.dirname(v)))
                        shutil.copyfile(os.path.join(cur_dir, v), os.path.join(new_dir, v))
                    elif not os.path.isabs(v):
                        p[k.split('.')[-1]] = os.path.join(cur_dir, v)

        if config_modifiers:
            s_out = s_out + config_modifiers

        s_out.save(strip_parents_keys=strip_parents_keys)

        return s_out

    def validate(self, module):
        try:
            assert module in self

            def val_fct(v, k, m=module):
                assert v, f'{m}.{k}={getattr(self, m)[k]} is invalid'
            getattr(self, f'validate_{module}')(val_fct)

        except AssertionError as error:
            print(f'Error: {error}')
            return False
        except ValueError as error:
            print(f'Error: {error}')
            return False

        return True



# File/Folder utilities
#
# Author: F. Mertens

import re
import os
import time
import shutil


def file_not_found_retry(fct, n_try, *args, **kargs):
    for i in range(n_try):
        try:
            fct(*args, **kargs)
            return 0
        except FileNotFoundError:
            if i + 1 == n_try:
                raise
            print('File not found error, will try again in a second ...')
            time.sleep(1 + i)


def reglob(path, exp, invert=False):
    m = re.compile(exp)

    if invert is False:
        res = [f for f in os.listdir(path) if m.search(f)]
    else:
        res = [f for f in os.listdir(path) if not m.search(f)]

    return map(lambda x: "%s/%s" % (path, x, ), res)


def rm_if_exist(directory, verbose=False):
    if os.path.exists(directory):
        if verbose:
            print('Remove ' + directory)
        shutil.rmtree(directory)


def mkdir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def abspath(p):
    return os.path.abspath(os.path.expanduser(p))


def is_file(f):
    return os.path.isfile(os.path.expanduser(f))


def zip_copy(dir_ins, dir_outs, filename, filename_out=None, verbose=True, n_retry=1):
    if not filename_out:
        filename_out = filename
    for dir_in, dir_out in zip(dir_ins, dir_outs):
        if verbose:
            print(f'Copy {dir_in}/{filename} -> {dir_out}/{filename_out}')
        file_not_found_retry(shutil.copyfile, n_retry, f'{dir_in}/{filename}', f'{dir_out}/{filename_out}')


def zip_copytree(dir_ins, dir_outs, dirname, verbose=True, n_retry=1):
    for dir_in, dir_out in zip(dir_ins, dir_outs):
        if verbose:
            print(f'Copy {dir_in}/{dirname} -> {dir_out}/{dirname}')
        file_not_found_retry(shutil.copytree, n_retry, f'{dir_in}/{dirname}', f'{dir_out}/{dirname}')


def zip_rename(dir_ins, dir_outs, filename_or_dirname=None, verbose=True, n_retry=1):
    for f_in, f_out in zip(dir_ins, dir_outs):
        if filename_or_dirname is not None:
            f_in = f'{f_in}/{filename_or_dirname}'
            f_out = f'{f_out}/{filename_or_dirname}'
        if verbose:
            print(f'Rename {f_in} -> {f_out}')
        file_not_found_retry(os.rename, n_retry, f_in, f_out)


def zip_rename_reg(dir_ins, dir_outs, regex, verbose=True, invert=False, n_retry=1):
    for f_in, f_out in zip(dir_ins, dir_outs):
        for f_in2 in reglob(f_in, regex, invert=invert):
            f_out2 = f'{f_out}/{os.path.basename(f_in2)}'
            if verbose:
                print(f'Rename {f_in2} -> {f_out2}')
            file_not_found_retry(os.rename, n_retry, f_in2, f_out2)


def zip_rm(dirs, verbose=True):
    for directory in dirs:
        rm_if_exist(directory, verbose=verbose)



# Some parts inspired by the AttrDict package written by Brendan Curran-Johnson (MIT Licence)

import re
import copy
import functools

from collections.abc import Mapping
import collections


def merge(left, right):
    """
    Merge two mappings objects together, combining overlapping Mappings,
    and favoring right-values
    left: The left Mapping object.
    right: The right (favored) Mapping object.
    NOTE: This is not commutative (merge(a,b) != merge(b,a)).
    """
    merged = collections.OrderedDict()

    left_keys = OrderedSet(left)
    right_keys = OrderedSet(right)

    # Items only in the left Mapping
    for key in left_keys - right_keys:
        merged[key] = left[key]

    # in both
    for key in left_keys & right_keys:
        left_value = left[key]
        right_value = right[key]

        if (isinstance(left_value, Mapping) and
                isinstance(right_value, Mapping)):  # recursive merge
            merged[key] = merge(left_value, right_value)
        else:  # overwrite with right value
            merged[key] = right_value

    # Items only in the right Mapping
    for key in right_keys - left_keys:
        merged[key] = right[key]

    return merged


def diff(left, right):
    d = collections.OrderedDict()

    left_keys = OrderedSet(left)
    right_keys = OrderedSet(right)

    # Items only in the left Mapping
    for key in left_keys - right_keys:
        d[key] = left[key]

    # in both
    for key in left_keys & right_keys:
        left_value = left[key]
        right_value = right[key]

        if (isinstance(left_value, Mapping) and
                isinstance(right_value, Mapping)):  # recursive merge
            diff_val = diff(left_value, right_value)
            if diff_val:
                d[key] = diff_val
        else:  # overwrite with right value
            if left_value == right_value:
                continue
            d[key] = left_value

    return d


def nested_dict_iter(nested):
    for key, value in nested.items():
        if isinstance(value, collections.abc.Mapping):
            for _, inner_key, inner_value in nested_dict_iter(value):
                yield value, f'{key}.{inner_key}', inner_value
        else:
            yield nested, key, value


class AttrMap(Mapping):
    """
    An implementation of MutableAttr.
    """

    def __init__(self, items=None):
        if items is None:
            items = collections.OrderedDict()
        elif not isinstance(items, Mapping):
            items = collections.OrderedDict(items)

        self._setattr('_mapping', items)

    def __getitem__(self, key):
        """
        Access a value associated with a key.
        """
        return self._mapping[key]

    def __setitem__(self, key, value):
        """
        Add a key-value pair to the instance.
        """
        self._mapping[key] = value

    def __delitem__(self, key):
        """
        Delete a key-value pair
        """
        del self._mapping[key]

    def __len__(self):
        """
        Check the length of the mapping.
        """
        return len(self._mapping)

    def __iter__(self):
        """
        Iterated through the keys.
        """
        return iter(self._mapping)

    def __repr__(self):
        """
        Return a string representation of the object.
        """
        # sequence type seems like more trouble than it is worth.
        # If people want full serialization, they can pickle, and in
        # 99% of cases, sequence_type won't change anyway
        return "AttrMap({mapping})".format(mapping=repr(self._mapping))

    @classmethod
    def _valid_name(cls, key):
        """
        Check whether a key is a valid attribute name.
        A key may be used as an attribute if:
         * It is a string
         * It matches /^[A-Za-z][A-Za-z0-9_]*$/ (i.e., a public attribute)
         * The key doesn't overlap with any class attributes (for Attr,
            those would be 'get', 'items', 'keys', 'values', 'mro', and
            'register').
        """
        return (
            isinstance(key, str) and
            re.match('^[A-Za-z][A-Za-z0-9_]*$', key) and
            not hasattr(cls, key)
        )

    def __getattr__(self, key):
        """
        Access an item as an attribute.
        """
        if key.startswith('_'):
            Mapping.__getattr__(self, key)

        if key not in self or not self._valid_name(key):
            raise AttributeError(
                "'{cls}' instance has no attribute '{name}'".format(
                    cls=self.__class__.__name__, name=key
                )
            )

        return self._build(self[key])

    def __add__(self, other):
        """
        Add a mapping to this Attr, creating a new, merged Attr.
        other: A mapping.
        NOTE: Addition is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        s = self.copy()
        s.set_data(merge(self, other))

        return s

    def __radd__(self, other):
        """
        Add this Attr to a mapping, creating a new, merged Attr.
        other: A mapping.
        NOTE: Addition is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        s = self.copy()
        s.set_data(merge(other, self))

        return s

    def __sub__(self, other):
        """
        Diff a mapping to this Attr, creating a new Attr.
        other: A mapping.
        NOTE: Addition is not commutative. a - b != b - a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        s = self.copy()
        s.set_data(diff(self, other))

        return s

    def __rsub__(self, other):
        """
        Diff a mapping to this Attr, creating a new Attr.
        other: A mapping.
        NOTE: Addition is not commutative. a - b != b - a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        s = self.copy()
        s.set_data(diff(other, self))

        return s

    def _build(self, obj):
        """
        Conditionally convert an object to allow for recursive mapping
        access.
        obj: An object that was a key-value pair in the mapping.
        """
        if isinstance(obj, Mapping):
            obj = AttrMap(obj)

        return obj

    def _setattr(self, key, value):
        """
        Add an attribute to the object, without attempting to add it as
        a key to the mapping.
        """
        Mapping.__setattr__(self, key, value)

    def __setattr__(self, key, value):
        """
        Add an attribute.
        key: The name of the attribute
        value: The attributes contents
        """
        if self._valid_name(key):
            self[key] = value
        elif key.startswith('_'):
            Mapping.__setattr__(self, key, value)
        else:
            raise TypeError(
                "'{cls}' does not allow attribute creation.".format(
                    cls=self.__class__.__name__
                )
            )

    def __delattr__(self, key, force=False):  # pragma: no cover
        """
        Delete an attribute.
        key: The name of the attribute
        """
        if self._valid_name(key):
            del self[key]
        elif key.startswith('_'):
            Mapping.__delattr__(self, key, force=force)
        else:
            raise TypeError(
                "'{cls}' does not allow attribute deletion.".format(
                    cls=self.__class__.__name__
                )
            )

    def set_data(self, mapping):
        self._mapping = mapping

    def copy(self):
        return copy.deepcopy(self)

    def update(self, other, allow_new=False):
        for key in other:
            if not allow_new:
                assert key in self, f'Key {key} does not exist'

            self[key] = other[key]

    def items(self):
        return self._mapping.items()

    def all_items(self):
        return nested_dict_iter(self)

    def all_keys(self):
        for _, k, _ in self.all_items():
            yield k

    def all_values(self):
        for _, _, v in self.all_items():
            yield v

    def get(self, key, default=None):
        try:
            return functools.reduce(lambda x, y: x[y], [self] + key.split('.'))
        except KeyError:
            return None


class OrderedSet(collections.abc.MutableSet):

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, nex = self.map.pop(key)
            prev[2] = nex
            nex[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['libpipe', 'libpipe.tests', 'libpipe.tools']

package_data = \
{'': ['*']}

install_requires = \
['asyncssh>=2.0,<3.0',
 'click>=7.0,<8.0',
 'progressbar2>=3.0,<4.0',
 'python-casacore>=3.0,<4.0',
 'tables>=3.2,<4.0',
 'toml>=0.10,<0.11']

entry_points = \
{'console_scripts': ['nodetool = libpipe.tools.nodetool:main']}

setup_kwargs = {
    'name': 'libpipe',
    'version': '0.1.2',
    'description': '',
    'long_description': 'Libpipe\n=======\n\nCommon pipeline framework and library\n\nInstallation\n------------\n\nlibpipe can be installed via pip:\n\n    $ pip install libpipe\n\nand requires Python 3.6.0 or higher.\n',
    'author': '"Florent Mertens"',
    'author_email': '"florent.mertens@gmail.com"',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/flomertens/libpipe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import os
import random

from unittest import mock

from libpipe import worker


def test_worker(tmp_path):
    w = worker.WorkerPool([worker.localhost_shortname, 'test_node'], 'Test', max_tasks_per_worker=5)

    assert len(w.workers) == 10
    assert isinstance(w.workers[0].client, worker.LocalClient)
    assert isinstance(w.workers[6].client, worker.Client)

    w = worker.WorkerPool([worker.localhost_shortname], 'Test')

    for i in range(20):
        w.add('touch %s/t%d' % (str(tmp_path), i))

    success, errors = w.execute()

    assert len(success) == 20
    assert len(errors) == 0

    for i in range(20):
        assert os.path.isfile('%s/t%d' % (str(tmp_path), i))


def test_worker_log(tmp_path):
    w = worker.WorkerPool(['wrong_server', worker.localhost_shortname], 'Test')

    for i in range(20):
        w.add(f'echo Task {i}', output_file=str(tmp_path / f't{i}.log'))
        print(str(tmp_path / f't{i}.log'))

    success, errors = w.execute()

    assert len(success) == 20
    assert len(errors) == 0

    for i in range(20):
        assert os.path.isfile(tmp_path / f't{i}.log')
        with open(tmp_path / f't{i}.log') as f:
            s = f.readlines()
        assert len(s) == 9 or len(s) == 7
        assert s[0].startswith('# Logging starting at ')
        assert s[1] == f'# Input command: echo Task {i}\n'
        assert s[-1].startswith('# Logging stopped at ')
        if len(s) == 7:
            assert s[3] == f'Task {i}\n'
        if len(s) == 9:
            assert s[3].startswith('Error')
            assert s[4] == f'Will retry 1 / 3 ...\n'
            assert s[5] == f'Task {i}\n'


def test_worker_wrong_server(tmp_path):
    w = worker.WorkerPool(['wrong_server'] * 6, 'Test')

    for i in range(20):
        w.add(f'echo Task {i}')

    success, errors = w.execute()

    assert len(success) == 0
    assert len(errors) == 20


def test_worker_env_file(tmp_path):
    with open(tmp_path / 'env', 'w') as f:
        f.write('export WORKER_TEST=test_env')

    w = worker.WorkerPool([worker.localhost_shortname], 'Test', env_source_file=str(tmp_path / 'env'))

    os.listdir(tmp_path)

    for i in range(20):
        w.add('echo env:${WORKER_TEST}; touch %s/t_${WORKER_TEST}_%d' % (str(tmp_path), i))

    success, errors = w.execute()

    assert len(success) == 20
    assert len(errors) == 0

    for i in range(20):
        assert os.path.isfile('%s/t_test_env_%d' % (str(tmp_path), i))


def test_worker_retcode():
    w = worker.WorkerPool([worker.localhost_shortname], 'Test')

    for i in range(10):
        w.add(f'echo Task {i}; exit {i}', name=f't{i}')

    success, errors = w.execute()

    assert len(success) == 10
    assert len(errors) == 0
    assert [f't{k.returncode}' for k in success] == [k.name for k in success]


def test_timeout():
    w = worker.WorkerPool([worker.localhost_shortname], 'Test', max_time=2)

    w.add('echo 0')

    success, errors = w.execute()

    assert len(success) == 1
    assert len(errors) == 0

    w = worker.WorkerPool([worker.localhost_shortname], 'Test', max_time=1)

    w.add('sleep 100')

    success, errors = w.execute()

    # Currently falling. Need to be fixed
    assert len(success) == 0
    assert len(errors) == 1


def test_done_callback():
    w = worker.WorkerPool([worker.localhost_shortname], 'Test', max_time=2)

    done = False

    def callback():
        nonlocal done
        done = True

    w.add('echo 0', done_callback=callback)

    success, errors = w.execute()

    assert len(success) == 1
    assert len(errors) == 0
    assert done


class FakeSSHClient(worker.LocalClient):

    def __init__(self, host, user=None, force_sync=True):
        worker.LocalClient.__init__(self, force_sync=force_sync)
        self.host = host

    async def execute(self, task):
        await worker.LocalClient.execute(self, task)
        task.remote_host = self.host


def test_run_on_host():
    with mock.patch.object(worker, 'Client', FakeSSHClient):
        n_hosts = 10
        hosts = [f'host{k}' for k in range(n_hosts)]
        w = worker.WorkerPool(hosts, 'Test', max_time=2)

        for i in range(20 * n_hosts):
            w.add('echo 0', run_on_host=random.choice(hosts))

        success, errors = w.execute()

        assert len(errors) == 0

        for task in success:
            assert task.remote_host == task.run_on_host
